/******************************************************

    Drawing of map, log and side bar in ASCII mode.

******************************************************/

#include "AUI.h"
#include "Creature.h"
#include "Hero.h"

static chtype FeatureTiles[shFeature::kMaxFeatureType];


void
shASCIIInterface::initializeGlyphs ()
{
    memset (mSqGlyphs, 0, sizeof (mSqGlyphs));

    mSqGlyphs[kStone] = ' ';
    mSqGlyphs[kVWall] = '#' | ColorMap[kBlue];
    mSqGlyphs[kHWall] = '#' | ColorMap[kBlue];
    mSqGlyphs[kVirtualWall] = '#' | ColorMap[kGreen];
    mSqGlyphs[kCavernWall] = '#' | ColorMap[kBrown];
    mSqGlyphs[kNWCorner] = '#' | ColorMap[kBlue];
    mSqGlyphs[kNECorner] = '#' | ColorMap[kBlue];
    mSqGlyphs[kSWCorner] = '#' | ColorMap[kBlue];
    mSqGlyphs[kSECorner] = '#' | ColorMap[kBlue];
    mSqGlyphs[kSewerWall] = '#' | ColorMap[kCyan];
    mSqGlyphs[kNTee] = '#' | ColorMap[kBlue];
    mSqGlyphs[kSTee] = '#' | ColorMap[kBlue];
    mSqGlyphs[kWTee] = '#' | ColorMap[kBlue];
    mSqGlyphs[kETee] = '#' | ColorMap[kBlue];

    mSqGlyphs[kFloor] = '.' | ColorMap[kBlue];
    mSqGlyphs[kBrokenLightAbove] = '.' | ColorMap[kBrightBlue];
    mSqGlyphs[kCavernFloor] = '.' | ColorMap[kBrown];
    mSqGlyphs[kSewerFloor] = '.' | ColorMap[kBrown];
    mSqGlyphs[kVirtualFloor] = '.' | ColorMap[kGreen];
    mSqGlyphs[kSewage] = '~' | ColorMap[kGreen];
    mSqGlyphs[kVoid] = '^' | ColorMap[kBlue];

    FeatureTiles[shFeature::kDoorHiddenVert] = '#' | ColorMap[kBlue];
    FeatureTiles[shFeature::kDoorHiddenHoriz] = '#' | ColorMap[kBlue];
    FeatureTiles[shFeature::kDoorClosed] = '+' | ColorMap[kCyan];
    FeatureTiles[shFeature::kMachinery] = '&' | ColorMap[kCyan];
    FeatureTiles[shFeature::kMovingHWall] = '=' | ColorMap[kCyan];
    FeatureTiles[shFeature::kStairsUp] = '<' | ColorMap[kWhite];
    FeatureTiles[shFeature::kStairsDown] = '>' | ColorMap[kWhite];
    FeatureTiles[shFeature::kVat] = '{' | ColorMap[kBrightGreen];
    FeatureTiles[shFeature::kComputerTerminal] = '_' | ColorMap[kGreen];
    FeatureTiles[shFeature::kPit] = '^' | ColorMap[kYellow];
    FeatureTiles[shFeature::kAcidPit] = '^' | ColorMap[kBrightYellow];
    FeatureTiles[shFeature::kSewagePit] = '^' | ColorMap[kGreen];
    FeatureTiles[shFeature::kRadTrap] = '^' | ColorMap[kBrightGreen];
    FeatureTiles[shFeature::kTrapDoor] = '^' | ColorMap[kRed];
    FeatureTiles[shFeature::kHole] = '^' | ColorMap[kGray];
    FeatureTiles[shFeature::kWeb] = '"' | ColorMap[kCyan];
    FeatureTiles[shFeature::kPortal] = '^' | ColorMap[kMagenta];
    FeatureTiles[shFeature::kPortableHole] = '#' | ColorMap[kRed];
    FeatureTiles[shFeature::kDoorOpen] = '\'' | ColorMap[kCyan];
}


void
shASCIIInterface::putGlyph (const shGlyph g)
{
    waddch (mMainWin, g.mSym | ColorMap[g.mColor]);
}


int /* Assumes cursor has been moved to right place in mMainWin. */
shASCIIInterface::drawSpecialEffect (shSpecialEffect e)
{
    static int spin_cnt = RNG (4);
    static int spin_dir = RNG (2) ? +1 : -1;
    switch (e) {
    case kNone:
    case kInvisibleEffect:
        return 0;
    case kExplosionEffect:
        waddch (mMainWin, '*' | ColorMap[kGray]);
        break;
    case kRadiationEffect:
        waddch (mMainWin, '*' | ColorMap[kGreen]);
        break;
    case kBlindingEffect:
        waddch (mMainWin, '*' | ColorMap[kWhite]);
        break;
    case kIncendiaryEffect:
        switch (RNG (3)) {
        case 0: waddch (mMainWin, '*' | ColorMap[kRed]); break;
        case 1: waddch (mMainWin, '*' | ColorMap[kBrightRed]); break;
        case 2: waddch (mMainWin, '*' | ColorMap[kBrightYellow]); break;
        }
        break;
    case kPsionicStormEffect: {
        char sym;
        shColor color;
        switch (RNG (4)) {
            case 0: sym = '/'; break;
            case 1: sym = '|'; break;
            case 2: sym = '-'; break;
            default: sym = '\\'; break;
        }
        switch (RNG (4)) {
            case 0: color = kBrightBlue; break;
            case 1: color = kBrightCyan; break;
            case 2: color = kBlue; break;
            default: color = kWhite; break;
        }
        waddch (mMainWin, sym | ColorMap[color]);
        break;
    }
    case kHeatEffect:
        waddch (mMainWin, '*' | ColorMap[kBrightRed]);
        break;
    case kColdEffect:
        waddch (mMainWin, '*' | ColorMap[kBlue]);
        break;
    case kPoisonEffect:
        waddch (mMainWin, '*' | ColorMap[kRed]);
        break;
    case kLaserBeamHorizEffect:
        waddch (mMainWin, '-' | ColorMap[kBrightCyan]);
        break;
    case kLaserBeamVertEffect:
        waddch (mMainWin, '|' | ColorMap[kBrightCyan]);
        break;
    case kLaserBeamBDiagEffect:
        waddch (mMainWin, '\\' | ColorMap[kBrightCyan]);
        break;
    case kLaserBeamFDiagEffect:
        waddch (mMainWin, '/' | ColorMap[kBrightCyan]);
        break;
    case kLaserBeamEffect:
        waddch (mMainWin, '*' | ColorMap[kBrightCyan]);
        break;
    case kRailHorizEffect:
        waddch (mMainWin, '-' | ColorMap[kBrightRed]);
        break;
    case kRailVertEffect:
        waddch (mMainWin, '|' | ColorMap[kBrightRed]);
        break;
    case kRailBDiagEffect:
        waddch (mMainWin, '\\' | ColorMap[kBrightRed]);
        break;
    case kRailFDiagEffect:
        waddch (mMainWin, '/' | ColorMap[kBrightRed]);
        break;
    case kRailEffect:
        //waddch (mMainWin, '*' | ColorMap[kBrightRed]);
        break;
    case kDisintegrationEffect:
        waddch (mMainWin, '*' | ColorMap[kBlue] | A_REVERSE);
        break;
    case kBinaryEffect:
        waddch (mMainWin, (RNG(2) ? '1' : '0') | ColorMap[kBrightGreen] | A_REVERSE);
        break;
    case kBugsEffect:
    {
        const char bugs[] = "~`@#$%^&*()-_=+:;'\"[]{}|\\/?>.<,";
        waddch (mMainWin, bugs[RNG(strlen(bugs))] | ColorMap[RNG(kBrightCyan)] | A_REVERSE);
        break;
    }
    case kVirusesEffect:
        waddch (mMainWin, '*' | ColorMap[kCyan]);
        break;
    case kSpitEffect:
        waddch (mMainWin, '.' | ColorMap[kGreen]);
        break;
    case kSpinEndEffect:
        spin_dir = RNG (2) ? +1 : -1;
        break;
    case kSpinEffect:
    {
        const char spin[] = "|/-\\";
        if (spin_dir > 0) {
            spin_cnt = ++spin_cnt > 3 ? 0 : spin_cnt;
        } else {
            spin_cnt = --spin_cnt < 0 ? 3 : spin_cnt;
        }
        waddch (mMainWin, spin[spin_cnt] | ColorMap[kGreen]);
        break;
    }

    case kRadarBlip:
        waddch (mMainWin, '0' | ColorMap[kRed]);
        break;
    case kSensedLife:
        waddch (mMainWin, 'I' | ColorMap[kYellow]);
        break;
    case kSensedTremor:
        waddch (mMainWin, '0' | ColorMap[kBrightRed]);
        break;

    default:
      return 0;
    }
    return 1;
}


void
shASCIIInterface::pSideWin (int y, int x, const char *fmt, ...)
{
    va_list ap;
    char linebuf[18];
    va_start (ap, fmt);
    vsnprintf (linebuf, 17, fmt, ap);
    va_end (ap);
    //linebuf[17] = '\0';
    wmove (mSideWin, y, x);
    waddstr (mSideWin, linebuf);
}


void
shASCIIInterface::setSideWinCol (shColor color)
{
    wattrset (mSideWin, ColorMap[color]);
}


void
shASCIIInterface::drawSideWinContent ()
{
    setSideWinCol (kWhite);
    pSideWin (0, 0, Hero.mName);

    setSideWinCol (kGray);
    pSideWin (1, 0, Hero.getTitle ());
    pSideWin (2, 0, "XL %2d:%d", Hero.mCLevel, Hero.getXP ());

    pSideWin (4, 0, "Str     Con    ");
    pSideWin (5, 0, "Agi     Dex    ");
    pSideWin (6, 0, "Int     Psi    ");

    FOR_ALL_ABILITIES (i) {
        if (i == kPsi && Hero.mPsiDrain) {
            setSideWinCol (Hero.mPsiDrain > 0 ? kRed : kGreen);
        } else if (Hero.mAbil.getByIndex (i) < Hero.mMaxAbil.getByIndex (i)) {
            setSideWinCol (kBrightYellow);
        } else {
            setSideWinCol (kGray);
        }
        pSideWin (4 + (i-1)/2, 12 - (i%2)*8, "%2d", Hero.mAbil.getByIndex (i));
    }

    setSideWinCol (kGray);
    pSideWin (8, 0, "Speed %+4d", Hero.mSpeed - 100);
    pSideWin (9, 0, "Armor  %3d", Hero.mAC);

    if (Hero.hasHealthMonitoring ()) { /* red < 1/6 >= yellow <= 1/2 > green */
        setSideWinCol (Hero.mHP < Hero.hpWarningThreshold () ? kBrightRed
                     : Hero.mHP <= Hero.mMaxHP / 2 ? kBrightYellow : kGreen);
    }
    pSideWin (10, 0, "HitPts %3d(%d)", Hero.mHP, Hero.mMaxHP);
    {
        int en, enmax;
        en = Hero.countEnergy (&enmax);
        setSideWinCol (kCyan);
        if (enmax) {
            pSideWin (11, 0, "Energy%4d(%d)", en, enmax);
        } else {
            pSideWin (11, 0, "Energy%4d", en);
        }
    }
    setSideWinCol (kGreen);
    pSideWin (12, 0, "$%d", Hero.countMoney ());

    shObject *weapon = NULL;
    if (Hero.mWeapon) {
        if (Hero.mWeapon->isA (kObjGenericComputer) && Hero.mCloak &&
            Hero.mCloak->isA (kObjPlasmaCaster))
        {
            weapon = Hero.mCloak;
        } else {
            weapon = Hero.mWeapon;
        }
    }
    if (weapon && weapon->isA (kWeapon)) {
        shObjectIlk *ilk = weapon->myIlk ();
        shObjectIlk *ammo = AllIlks[ilk->mAmmoType];

        if (ammo) {
            int n = 0;
            for (int i = 0; i < Hero.mInventory->count (); ++i) {
                shObject *obj = Hero.mInventory->get (i);
                if (obj->isA (ammo->mId)) {
                    n += obj->mCount;
                }
            }
            setSideWinCol (ammo->mGlyph.mColor);
            pSideWin (12, 7, "%c", ammo->mGlyph.mSym);
            setSideWinCol (kGray);
            pSideWin (12, 8, "%d", n);
        }
    } else if (weapon && weapon->isA (kRayGun) && weapon->isChargeKnown ()) {
        setSideWinCol (kGray);
        pSideWin (12, 7, "(%d)", weapon->mCharges);
    }
    { /* Conditions */
        int n = 13;
        const char *condition = NULL;
        setSideWinCol (kBrown);
        /* Sort them in order of most likely to be temporary first. */
        /* There is only space for six. */

        if (Hero.is (kUnableCompute) || Hero.is (kUnableUsePsi) ||
            Hero.is (kUnableAttack))
        {
            pSideWin (n++, 0, "Commanded");
        }
        if (Hero.is (kFrightened)) {
            pSideWin (n++, 0, "Frightened");
        }
        if (Hero.is (kStunned)) {
            pSideWin (n++, 0, "Stunned");
        }
        if (Hero.is (kConfused)) {
            pSideWin (n++, 0, "Confused");
        }
        if (Hero.isInvisible ()) {
            pSideWin (n++, 0, "Cloaked");
        } else if (Hero.hasCamouflage ()) {
            pSideWin (n++, 0, "Camouflaged");
        }
        if (Hero.is (kHosed)) {
            pSideWin (n++, 0, "Hosed");
        }
        if (Hero.is (kSickened)) {
            pSideWin (n++, 0, "Sickened");
        }
        if (Hero.is (kViolated)) {
            pSideWin (n++, 0, "Sore");
        }
        if (Hero.getStoryFlag ("impregnation") > 0) {
            pSideWin (n++, 0, "Pregnant");
        }
        if (Hero.is (kPlagued)) {
            pSideWin (n++, 0, "PLAGUED");
        }
        if (Hero.getStoryFlag ("superglued tongue")) {
            pSideWin (n++, 0, "Mute");
        }
        int radsymptom = Hero.getStoryFlag ("radsymptom");
        if (radsymptom > 1) {
            pSideWin (n++, 0, radsymptom > 18 ? "Radiation Ghoul" :
                              radsymptom >  9 ? "Radiation Zombie" :
                                                "Radiation Sick");
        }
        if (Hero.isBlind ()) { /* Is obvious so goes last. */
            pSideWin (n++, 0, "Blind");
        }

        switch (Hero.getEncumbrance ()) {
        case kOverloaded: condition = "Overloaded"; break;
        case kOvertaxed: condition = "Overtaxed"; break;
        case kStrained: condition = "Strained"; break;
        case kBurdened: condition = "Burdened";
        }
        if (condition) {
            pSideWin (n++, 0, condition);
        }
    }
    setSideWinCol (kWhite);
    pSideWin (GodMode ? 17 : 19, 0, "%5s %d", Level->mName, Level->mDLevel);
    if (GodMode) {
        setSideWinCol (kGray);
        //pSideWin (17, 0, "wt %5d/%5d", Hero.mWeight, Hero.mCarryingCapacity);
        pSideWin (18, 0, "rad %d ", Hero.mRad);
        pSideWin (19, 0, "%d ", Clock);
    }
}


void
shASCIIInterface::drawSideWin ()
{
    if (!Hero.mProfession) { return; }
    werase (mSideWin);
    drawSideWinContent ();
    wnoutrefresh (mSideWin);
}


void
shASCIIInterface::drawLog ()
{
    if (mDiagWin) {
        touchwin (mDiagWin);
        wnoutrefresh (mDiagWin);
    }
    touchwin (mLogWin);
    wnoutrefresh (mLogWin);
    doupdate ();
}


void
shASCIIInterface::drawScreen ()
{
    Level->draw ();
    drawSideWin ();
    wnoutrefresh (mMainWin);
    drawLog (); /* calls doupdate() for us */
}


void
shASCIIInterface::refreshScreen ()
{
    wnoutrefresh (mMainWin);
    drawLog ();
}


/* Returns glyph of a feature. Handles fancy coloring of doors. */
chtype      /* Does not expect NULL pointer. */
shASCIIInterface::featureGlyph (shFeature *f)
{
    if (f->isDoor () && !f->isHiddenDoor ()) {
        chtype glyph;
        if (!f->isOpenDoor ()) {
            glyph = f->isMagneticallySealed () ? '|' :
                    (f->isBerserkDoor () && !f->mTrapUnknown) ? '"' : '+';
        } else {
            glyph = (f->isBerserkDoor () && !f->mTrapUnknown) ? '`' : '\'';
        }
        if (f->isLockDoor ()) {
            if (f->isLockBrokenDoor ()) {
                if (f->isLockedDoor ()) { /* Damaged and jammed lock. */
                    glyph |= ColorMap[kBrightCyan];
                } else { /* Badly broken lock.  Can be still repaired. */
                    glyph |= ColorMap[kGray];
                }
            } else if (f->isRetinaDoor ()) {
                glyph |= ColorMap[kWhite];
            } else if (f->isCodeLockDoor ()) {
                shObjectIlk *ilk = f->keyNeededForDoor ();
                if (ilk) { /* Door color matches keycard color. */
                    glyph |= ColorMap[ilk->mGlyph.mColor];
                } else { /* Lock color unknown. */
                    glyph |= ColorMap[kBrown];
                }
            } else { /* Something not right. */
                glyph |= ColorMap[kMagenta];
            }
        } else { /* Door without lock. */
            glyph |= ColorMap[kCyan];
        }
        return glyph;
    } else {
        return FeatureTiles[f->mType];
    }
}

/* In ASCII mode only the most important thing on a tile is displayed. */
void
shASCIIInterface::draw (int x, int y, shSpecialEffect e, shCreature *c,
                   shObjectIlk *o, shFeature *f, shSquare *s)
{
    wmove (mMainWin, y - mX0, x - mX0);

    if (e) { /* Explosions and the like obscure everything. */
        if (drawSpecialEffect (e)) return;
    }

    if (c) { /* Creatures have highest priority. */
        if (c->isHero () && Hero.mBodyArmor &&
            Hero.mBodyArmor->isPoweredArmor () &&
            Hero.mBodyArmor->isAppearanceKnown ())
        {   /* Powered body armor changes hero glyph color. */
            waddch (mMainWin, c->mGlyph.mSym |
                ColorMap[Hero.mBodyArmor->myIlk ()->mGlyph.mColor]);
        } else if (c->isPet ()) {
            shGlyph g = c->mGlyph;
            waddch (mMainWin, g.mSym | ColorMap[g.mColor] | A_REVERSE);
        } else {
            putGlyph (c->mGlyph);
        }
        return;
    }

    if (o) { /* Next is object on top of a pile. */
        putGlyph (o->mGlyph);
        return;
    }

    if (f) { /* Features unobscured by objects. */
        waddch (mMainWin, featureGlyph (f));
        return;
    }

    if (s) { /* Finally, terrain. */
        if (Level->isFloor (x, y) && /* Flashlight radius. */
            Hero.hasLightSource () && distance (&Hero, x, y) <= 25)
        {
            chtype light = (mSqGlyphs[s->mTerr] & A_CHARTEXT) | ColorMap[kBrightYellow];
            waddch (mMainWin, light);
        } else {
            waddch (mMainWin, mSqGlyphs[s->mTerr]);
        }
        return;
    }
    waddch (mMainWin, ' ');
}


void
shASCIIInterface::drawMem (int x, int y, shSpecialEffect e, shCreature *c,
                      shObjectIlk *o, shFeature *f, shSquare *s)
{
    shMapLevel::shMemory mem = Level->getMemory (x, y);
    wmove (mMainWin, y - mX0, x - mX0);

    if (e) {
        if (drawSpecialEffect (e)) return;
    }

    if (c) {
        if (c->isHero () && Hero.mBodyArmor &&
            Hero.mBodyArmor->isPoweredArmor () &&
            Hero.mBodyArmor->isAppearanceKnown ())
        {
            waddch (mMainWin, c->mGlyph.mSym |
                ColorMap[Hero.mBodyArmor->myIlk ()->mGlyph.mColor]);
        } else {
            putGlyph (c->mGlyph);
        }
        return;
    } else if (mem.mMon) {
        putGlyph (MonIlks[mem.mMon]->mGlyph);
        return;
    }

    if (o) {
        putGlyph (o->mGlyph);
        return;
    } else if (mem.mObj) {
        putGlyph (AllIlks[mem.mObj]->mGlyph);
        return;
    }

    if (f) {
        waddch (mMainWin, featureGlyph (f));
        return;
    } else if (mem.mFeat != shFeature::kMaxFeatureType) {
        /* Ugly kludge here. Only door flags are saved, not whether
           door malfunctions are known by hero. */
        shFeature *real = Level->getFeature (x, y);
        shFeature fake;
        fake.mType = mem.mFeat;
        fake.mDoor.mFlags = mem.mDoor;
        if (real)
            fake.mTrapUnknown = real->mTrapUnknown;
        waddch (mMainWin, featureGlyph (&fake));
        return;
    }

    if (s) { /* Finally, terrain. */
        if (Hero.mZ < 0) { /* In a pit. */
            if (Level->isFloor (x, y)) {
                waddch (mMainWin, ' ');
            } else {
                waddch (mMainWin, mSqGlyphs[s->mTerr]);
            }
        } else {
            if (Flags.mShowLOS && Level->isFloor (x, y)) {
                waddch (mMainWin, ' ');
            } else {
                waddch (mMainWin, mSqGlyphs[s->mTerr]);
            }
        }
        return;
    } else if (mem.mTerr != kMaxTerrainType) {
        if (Hero.mZ < 0 && Level->isFloor (x, y)) {
            waddch (mMainWin, ' ');
        } else {
            waddch (mMainWin, mSqGlyphs[mem.mTerr]);
        }
        return;
    }
    waddch (mMainWin, ' ');
}
