#include <ctype.h>
#include "Interface.h"
#include "Hero.h"

void
shInterface::cursorOnHero ()
{
    cursorOnXY (Hero.mX, Hero.mY);
}

void
shInterface::pause ()
{
    doMorePrompt ();
}

void
shInterface::nevermind ()
{
    p ("Never mind.");
}

void
shInterface::smallPause ()
{
    mPause = 2;
}

void
shInterface::nonlOnce ()
{
    mNoNewline = 1;
}

int
shInterface::freeLogLines ()
{
    return mLogSize - mLogSCount;
}


shInterface::Command
shInterface::getCommand ()
{
    mLogSCount = 0;
    cursorOnHero ();
    int res = getChar ();
    pageLog ();
    return keyToCommand (res);
}


int
shInterface::p (const char *format, ...)
{
    int res;
    va_list ap;
    va_start (ap, format);
    res = vp (format, ap);
    va_end (ap);
    return res;
}


int
shInterface::vp (const char *format, va_list ap)
{
    const int buflen = 80;
    char strbuf[buflen];
    int res;

    if (6 == ++mLogSCount || mPause) {
        doMorePrompt ();
        mLogSCount = 1;
    }

    /* Capitalization rules. */
    res = vsnprintf (&strbuf[0], buflen, format, ap);
    if (' ' != strbuf[1] || '-' != strbuf[2]) {
        /* For "c - pea shooter" messages. */
        strbuf[0] = toupper (strbuf[0]);
    }
    if ('"' == strbuf[0]) {
        strbuf[1] = toupper (strbuf[1]);
    }
    strbuf[buflen-1] = 0;
    debug.log ("%s", strbuf);

    int newcol;
    if (mNoNewline) {
        --mHistoryIdx;
        if (mHistoryIdx < 0) mHistoryIdx = HISTORY_ROWS - 1;
        strcpy (&mLogHistory[mHistoryIdx*80] + mLogCol, strbuf);
        newcol = mLogCol + strlen (strbuf);
    } else {
        strcpy (&mLogHistory[mHistoryIdx*80], strbuf);
        newcol = strlen (strbuf);
    }
    mHistoryIdx = (mHistoryIdx + 1) % HISTORY_ROWS;
    if (0 == mHistoryIdx && !mNoNewline) {
        mHistoryWrapped = 1;
    }
    int nonl = mNoNewline;

    logPrint (strbuf);
    if (!nonl && mLogRow < mLogSize) {
        mLogRow++;
    }
    mLogCol = newcol;
    drawLog ();
    return res;
}


int
shInterface::yn (const char *prompt, ...)
{
    char msg[80];
    int res;

    snprintf (msg, 80, "%s [y/n]", prompt);
    msg[79] = 0;

    va_list ap;
    va_start (ap, prompt);
    vp (msg, ap);
    va_end (ap);

    while (1) {
        res = getChar ();
        if ('y' == res || 'Y' == res) {
            return 1;
        } else if ('n' == res || 'N' == res) {
            return 0;
        }
    }
}


void
shInterface::pauseXY (int x, int y, int ms /* = 0 */, int draw /* = 1 */)
{
    if (ms) {
        if (draw) drawScreen ();
    } else {
        moreBegin ();
    }
    Level->drawSq (x, y);
    cursorOnXY (x, y);
    if (ms) { /* Wait specified time. */
        delay (ms);
    } else {  /* Wait until keypress. (space/enter/esc) */
        int ch;
        do {
            ch = getChar ();
        } while (' ' != ch && 13 != ch && 10 != ch && 27 != ch);
    }
    if (!ms) {
        moreEnd ();
        mLogSCount = 0;
    }
}


void
shInterface::doMorePrompt ()
{
    moreBegin ();
    int ch;
    do {
        ch = getChar ();
    } while (' ' != ch && 13 != ch && 10 != ch && 27 != ch);

    if (2 != mPause) {
        mLogSCount = 0;
        pageLog ();
    }
    mPause = 0;
    moreEnd ();
    if (Flags.mFadeLog) drawLog ();
}


shDirection
shInterface::getDirection (int *x, int *y, int *z, int silent)
{
    int res;
    if (!silent) {
        pageLog ();
        p ("In what direction?");
    }
    drawLog ();
    cursorOnHero ();
    do {
        res = getChar ();
        if (z) *z = 0;
        Command cmd;
        if (' ' == res) {
            cmd = kSpace;
        } else if (27 == res) {
            cmd = kEscape;
        } else if (13 == res || 10 == res) {
            cmd = kEnter;
        } else {
            cmd = keyToCommand (res);
        }
        switch (cmd) {
        case kMoveSW: if (x) --*x; if (y) ++*y; pageLog (); return kSouthWest;
        case kMoveS:  if (y) ++*y; pageLog (); return kSouth;
        case kMoveSE: if (x) ++*x; if (y) ++*y; pageLog (); return kSouthEast;
        case kMoveW:  if (x) --*x; pageLog (); return kWest;
        case kRest: case kGlide: pageLog (); return kOrigin;
        case kMoveE:  if (x) ++*x; pageLog (); return kEast;
        case kMoveNW: if (x) --*x; if (y) --*y; pageLog (); return kNorthWest;
        case kMoveN:  if (y) --*y; pageLog (); return kNorth;
        case kMoveNE: if (x) ++*x; if (y) --*y; pageLog (); return kNorthEast;
        case kMoveUp: if (z) --*z; pageLog (); return kUp;
        case kMoveDown: if (z) ++*z; pageLog (); return kDown;
        case kSpace: case kEnter: case kEscape:
            pageLog ();
            nevermind ();
            return kNoDirection;
        default:
            pageLog ();
            p ("NW: %s N: %s NE: %s       up: %s",
                getKeyForCommand (kMoveNW), getKeyForCommand (kMoveN),
                getKeyForCommand (kMoveNE), getKeyForCommand (kMoveUp));
            p (" W: %s       E: %s     self: %s",
                getKeyForCommand (kMoveW), getKeyForCommand (kMoveE),
                getKeyForCommand (kRest));
            p ("SW: %s S: %s SE: %s     down: %s",
                getKeyForCommand (kMoveSW), getKeyForCommand (kMoveS),
                getKeyForCommand (kMoveSE), getKeyForCommand (kMoveDown));
            p ("Please select direction from above possibilities.");
            p ("Hit space, enter or esacape to cancel.");
            cursorOnHero ();
            break;
        }
    } while (1);
}

/*maxradius is specified in squares
  returns: 1 on success, 0 on abort
*/
int
shInterface::getSquare (const char *prompt, int *x, int *y, int maxradius,
                        int instant)
{
    if (maxradius <= 0) maxradius = 99;
    if (prompt) p (prompt);

    if (-1 == *x || -1 == *y) {
        *x = Hero.mX;
        *y = Hero.mY;
    }
    int tx, ty;
    int step = 1;

    while (1) {
        cursorOnXY (*x, *y);
        tx = *x; ty = *y;
        int key = getChar ();
        Command cmd;
        if (' ' == key) {
            cmd = kSpace;
        } else if (27 == key) {
            cmd = kEscape;
        } else if (13 == key) {
            cmd = kEnter;
        } else {
            cmd = keyToCommand (key);
        }
        switch (cmd) {
        case kMoveSW: tx = *x - step; ty = *y + step; break;
        case kMoveS:                  ty = *y + step; break;
        case kMoveSE: tx = *x + step; ty = *y + step; break;
        case kMoveW:  tx = *x - step;                 break;
        case kMoveE:  tx = *x + step;                 break;
        case kMoveNW: tx = *x - step; ty = *y - step; break;
        case kMoveN:                  ty = *y - step; break;
        case kMoveNE: tx = *x + step; ty = *y - step; break;
        case kRest: case kSpace: case kEnter:
            return 1;
        case kEscape:
            return 0;
        case kHelp:
            if (Level->isInBounds (*x, *y) &&
                Level->isOccupied (*x, *y))
            {
                shCreature *c = Level->getCreature (*x, *y);
                if (Hero.canSee (c) || Hero.canHearThoughts (c))
                    c->showLore ();
            }
            continue;
        default:
            I->p ("Use the movement keys to position the cursor.");
            I->p ("Confirm the location with '%s' key, space or enter.",
                  I->getKeyForCommand (kRest));
            I->p ("Press escape to abort.  Press '%s' for creature lore.",
                  I->getKeyForCommand (kHelp));
            continue;
        }
        if (Level->isInBounds (tx, ty) &&
            distance (&Hero, tx, ty) <= 5 * maxradius)
        {
            *x = tx;
            *y = ty;
        }
        if (instant)
            return 1;
    }
}


void
shInterface::readKeybindings (const char *fname, int n_keys, const shNameToKey KeyNames[])
{
    static const struct shCommandName
    {
        const char *name;
        Command cmd;
    } CommandNames[] =
    {
        {"move n", kMoveN},
        {"move ne", kMoveNE},
        {"move e", kMoveE},
        {"move se", kMoveSE},
        {"move s", kMoveS},
        {"move sw", kMoveSW},
        {"move w", kMoveW},
        {"move nw", kMoveNW},
        {"move down", kMoveDown},
        {"move up", kMoveUp},

        {"run n", kGlideN},
        {"run ne", kGlideNE},
        {"run e", kGlideE},
        {"run se", kGlideSE},
        {"run s", kGlideS},
        {"run sw", kGlideSW},
        {"run w", kGlideW},
        {"run nw", kGlideNW},
        {"run down", kGlideDown},
        {"run up", kGlideUp},

        {"shoot n", kFireN},
        {"shoot ne", kFireNE},
        {"shoot e", kFireE},
        {"shoot se", kFireSE},
        {"shoot s", kFireS},
        {"shoot sw", kFireSW},
        {"shoot w", kFireW},
        {"shoot nw", kFireNW},
        {"shoot down", kFireDown},
        {"shoot up", kFireUp},

        {"adjust inventory", kAdjust},
        {"close", kClose},
        {"discoveries", kDiscoveries},
        {"drop", kDrop},
        {"multi drop", kDropMany},
        {"skills", kEditSkills},
        {"examine", kExamine},
        {"execute", kExecute},
        {"shoot", kFireWeapon},
        {"run", kGlide},
        {"help", kHelp},
        {"show message history", kHistory},
        {"kick", kKick},
        {"show inventory", kListInventory},
        {"look around", kLook},
        {"use mutant power", kMutantPower},
        {"name", kName},
        {"open", kOpen},
        {"pay", kPay},
        {"pick up", kPickup},
        {"pick up all", kPickUpAll},
        {"quaff", kQuaff},
        {"rest", kRest},
        {"take off", kTakeOff},
        {"toggle autopickup", kToggleAutopickup},
        {"throw", kThrow},
        {"apply", kUse},
        {"wear", kWear},
        {"wield", kWield},
        {"zap ray gun", kZapRayGun},
        {"debug command", kGodMode}
    };
    const int n_commands = sizeof (CommandNames) / sizeof (shCommandName);

    char keymapname[PRIME_PATH_LENGTH];
    snprintf (keymapname, sizeof (keymapname)-1, "%s/%s", DATADIR, fname);
    FILE *keymap = fopen (keymapname, "r");
    if (!keymap) {
        p ("Could not read '%s'.", keymapname);
        return; /* Loading default keymap is in main.cpp. */
    } /* Save keymap file name. */
    strncpy (&Flags.mKeySet[0], fname, sizeof (Flags.mKeySet)-1);
    resetKeybindings ();
    int scan, errors = 0;
    do {
        char buf[200];
        char *cmd_str, *key_str;
        Command command;
        int key;
        buf[0] = 0;
        scan = NULL != fgets (&buf[0], sizeof (buf)-1, keymap);
        if (buf[0] == '\n' || buf[0] == 0) { /* Blank line. */
            continue;
        } else if (buf[0] == '#') { /* Comment character. */
            continue;
        } /* Parse 'command = key' style assignments. */
        cmd_str = &buf[0];
        key_str = strchr (&buf[0], '=');
        if (!key_str) {
            debug.log ("Expected '=' in key to command assignment:");
            debug.log ("%s", buf);
            ++errors;
            continue;
        } /* Remove whitespace from beginnings. */
        while (isspace (*cmd_str))
            ++cmd_str;
        if (*cmd_str == '=') {
            debug.log ("Invalid null command assignment:");
            debug.log ("%s", buf);
            ++errors;
            continue;
        }
        ++key_str;
        while (isspace (*key_str))
            ++key_str;
        if (*key_str == '\0') {
            debug.log ("Invalid null key assignment:");
            debug.log ("%s", buf);
            ++errors;
            continue;
        } /* Remove whitespace from endings. */
        char *work = strchr (&buf[0], '=') - 1;
        while (isspace (*work))
            *(work--) = 0;
        work = key_str + strlen (key_str) - 1;
        while (isspace (*work))
            *(work--) = 0;
        /* Ignore case. */
        for (int i = strlen (cmd_str) - 1; i >= 0; --i) {
            cmd_str[i] = tolower (cmd_str[i]);
        } /* Convert only if not letter (with control char). */
        int key_strlen = strlen (key_str);
        if (key_strlen > 2) {
            for (int i = strlen (key_str) - 1; i >= 0; --i) {
                key_str[i] = tolower (key_str[i]);
            }
        } /* Find command name in array. */
        int n = 0;
        while (n < n_commands && strcmp (CommandNames[n].name, cmd_str))
            ++n;
        if (n == n_commands) { /* Skip this line and continue. */
            debug.log ("Unknown command '%s'.", cmd_str);
            ++errors;
            continue;
        } /* Success. */
        command = CommandNames[n].cmd;
        if (key_strlen > 2) { /* Find key name in array. */
            int n = 0;
            while (n < n_keys && strcmp (KeyNames[n].name, key_str))
                ++n;
            if (n == n_keys) {
                debug.log ("Unknown key '%s'.", key_str);
                ++errors;
                continue;
            }
            key = KeyNames[n].key;
        } else {
            if (key_strlen == 2) {
                if (key_str[0] != '^' && key_str[0] != '~') {
                    debug.log ("Expected modifier character: %s.", key_str);
                    ++errors;
                    continue;
                } else if (key_str[0] == '^') {
                    key = 0x1f & key_str[1]; /* Control key. */
                } else {
                    key = 0x80 | key_str[1]; /* Alt key. */
                }
            } else {
                key = key_str[0];
            }
        } /* Success. Proceed to assignment. */
        debug.log ("Assigning '%c'(%d) to %s.", key, key, CommandNames[n].name);
        assign (key, command);
    } while (scan);
    assign (27, kEscape); /* This will make sure it is not overridable. */
    if (errors) {
        p ("There %s %d parse error%s reading %s.",
            errors > 1 ? "were" : "was", errors,
            errors > 1 ? "s" : "", keymapname);
    }
    fclose (keymap);
}


void
shMapLevel::draw ()
{
    for (int y = 0; y < MAPMAXROWS; ++y) {
        for (int x = 0; x < MAPMAXCOLUMNS; ++x) {
            drawSq (x, y);
        }
    }
}


int
shMapLevel::rememberedCreature (int x, int y)
{
    return mRemembered[x][y].mMon != kMonEarthling;
}

void
shMapLevel::feelSq (int x, int y)
{
    remember (x, y, getSquare (x, y)->mTerr);
    shFeature *f = getKnownFeature (x, y);
    if (f) {
        if (f->isDoor ()) {
            /* Feeling the door does not reveal what color is door lock. */
            mRemembered[x][y].mFeat = f->mType;
            /* On the other hand lock type can be checked easily. */
            mRemembered[x][y].mDoor |=
                (f->mDoor.mFlags & shFeature::kCanFeel);
        } else {
            remember (x, y, f);
        }
    }
    if (isOccupied (x, y) && (Hero.mX != x || Hero.mY != y) &&
        !Hero.isAwareOf (getCreature (x, y)))
    {
        remember (x, y, kMonInvisible);
    } else {
        forgetCreature (x, y);
    }
    if (countObjects (x, y) > 0) {
        shObjectVector *objs = getObjects (x, y);
        shObject *bestobj = findBestObject (objs);
        remember (x, y, bestobj->mIlkId);
    } else {
        forgetObject (x, y);
    }
    Level->drawSq (x, y);
}

void
shMapLevel::drawSq (int x, int y, int forget /* = 0 */)
{
    shCreature *c = getCreature (x, y);
    shFeature *f = getFeature (x, y);
    shFeature *kf = getKnownFeature (x, y);
    shSpecialEffect e = getSpecialEffect (x, y);
    if (!Hero.canSee (x, y)) {
        /* Explosions and beams light up dark area for a moment. */
        if (!Hero.isBlind () && isInLOS (x, y) && e) {
            I->draw (x, y, e, NULL, NULL, NULL, NULL);
            return;
        }
        if (c == &Hero) { /* You know where you are.  Duh. */
            I->drawMem (x, y, kNone, c, NULL, NULL, NULL);
        } else if (c && Hero.canHearThoughts (c)) {
            forgetCreature (x, y);
            I->drawMem (x, y, kNone, c, NULL, NULL, NULL);
        } else if (c && (Hero.hasNightVision () || Hero.hasEMFieldVision ()) &&
                   Hero.canSee (c))
        {
            forgetCreature (x, y);
            I->drawMem (x, y, kNone, c, NULL, NULL, NULL);
        } else if (c && Hero.hasSenseLife () && Hero.canSenseLife (c)) {
            I->drawMem (x, y, kSensedLife, NULL, NULL, NULL, NULL);
        } else if (Hero.hasMotionDetection () && !Hero.isBlind () &&
            (    /* Moving creature. */
             (c && c->isMoving ())
              || /* Moving automatic door. */
             (f && f->isDoor () && f->mDoor.mMoveTime + SLOWTURN >= Clock)
            ) && distance (&Hero, x, y) < 50)
        {
            I->drawMem (x, y, kRadarBlip, NULL, NULL, NULL, NULL);
            if (!(AllIlks[kObjMotionTracker]->mFlags & kIdentified)) {
                AllIlks[kObjMotionTracker]->mFlags |= kIdentified;
                I->p ("You now recognize motion trackers.");
            }
        } else if (c && Hero.hasTremorsense () && Hero.canFeelSteps (c)) {
            I->drawMem (x, y, kSensedTremor, NULL, NULL, NULL, NULL);
        } else if (Hero.hasNightVision () && isInLOS (x, y) &&
                   kf && kf->isDoor () && kf->isOpenDoor () &&
                   Level->getMemory (x, y).mFeat == shFeature::kDoorClosed)
        { /* Kludge: show doors because, e.g. night vision might have
             revealed a creature on the other side that hero might not
             expect to see if she remembers the door was closed. */
            I->drawMem (x, y, kNone, NULL, NULL, kf, NULL);
        } else { /* Everything from memory. */
            I->drawMem (x, y);
        }
        return;
    }

    shObjectVector *objs = getObjects (x, y);
    shObject *bestobj = NULL;
    shObjectIlk *objilk = NULL;
    if (objs && !Level->isWatery (x, y) && (bestobj = findBestObject (objs))) {
        objilk = bestobj->myIlk ();
        Level->remember (x, y, objilk->mId);
    } else {
        Level->forgetObject (x, y);
    }

    if (c) { /* Creatures may stay in hiding. */
        if (c->mHidden > 0 && Hero.canHearThoughts (c)) {
            c->mHidden *= -1; /* Show the hidden creature from now on. */
        }
        if (c->mHidden > 0) { /* Hero is not aware of the monster. */
            switch (c->mMimic) {
            case shCreature::kObject:
                Level->remember (x, y, c->mMimickedObject->mId);
                objilk = c->mMimickedObject; /* Obscure items. */
                break;
            default:
                break; /* Invisible, or hiding under existing object. */
            }
            c = NULL;
        }
        if (c && !Hero.canSee (c)) {
            if (!e && Hero.hasMotionDetection () && !Hero.isBlind () &&
                c->isMoving () && distance (&Hero, x, y) < 50)
            {
                e = kRadarBlip;
            }
            c = NULL;
        }
        if (c) {
            if (c->isSessile ()) { /* Mark non-moving monsters. */
                Level->remember (x, y, c->mIlkId);
            } else { /* Clear any remembered creature leftovers. */
                forgetCreature (x, y);
            }
        }
    }
    if (!c) forgetCreature (x, y);

    if (f && f->isHiddenDoor ()) { kf = f; } /* Drawn as walls. */
    if (kf) {
        Level->remember (x, y, kf);
    } else {
        Level->forgetFeature (x, y);
    }

    shSquare *s = getSquare (x, y);
    if (isFloor (x, y) && !isLit (x, y, x, y) && !Hero.isBlind ()) {
        Level->forgetTerrain (x, y);
    } else {
        Level->remember (x, y, s->mTerr);
    }

    I->draw (x, y, e, c, objilk, kf, s);
}


void
initGlyph (shGlyph *g, char sym, shColor color)
{
    g->mSym = sym;
    g->mColor = color;
}

#ifndef _WIN32
#include <time.h>   // nanosleep

void
shInterface::delay (int ms)
{
    timespec wait;
    wait.tv_sec = ms / 1000;
    wait.tv_nsec = (ms % 1000) * 1000000; /* Nanoseconds. */
    nanosleep (&wait, NULL);
}
#else
#undef DATADIR
#include <windows.h>

void
shInterface::delay (int ms)
{
    Sleep (ms);
}
#endif
