#include <ctype.h>
#include <math.h>
#include "Global.h"
#include "Object.h"
#include "Interface.h"
#include "ObjectType.h"
#include "Creature.h"
#include "Hero.h"
#include "OperatingSystem.h"

shGlyph ObjectGlyphs[kMaxObjectType];

char ObjectSymbols[] = { '9', '$', ':', '?', '!', '(', '[',
                         ')', '=', '&', '/' };


shObjectDescData JumpsuitData[30] =
{
    { "plaid jumpsuit", kGreen },
    { "yellow jumpsuit", kYellow },
    { "green jumpsuit", kBrightGreen },
    { "blue jumpsuit", kBrightBlue },
    { "black jumpsuit", kBlue }             //  5
};

shObjectDescData RayGunData[30] =
{
    { "yellow ray gun", kYellow },
    { "indigo ray gun", kBlue },
    { "blue ray gun", kBrightBlue },
    { "red ray gun", kRed },
    { "cyan ray gun", kCyan },              //  5
    { "magenta ray gun", kMagenta },
    { "white ray gun", kWhite },
    { "green ray gun", kGreen },
    { "orange ray gun", kBrightRed },
    { "silver ray gun", kBrightCyan },      // 10
    { "gold ray gun", kYellow },
    { "purple ray gun", kMagenta }
};

shObjectDescData ImplantData[30] =
{
    { "nodular bionic implant", kGray },
    { "slippery bionic implant", kGreen },
    { "pulsating bionic implant", kBrightGreen },
    { "quivering bionic implant", kMagenta },
    { "stretchy bionic implant", kRed },            //  5
    { "perforated bionic implant", kBrightBlue },
    { "squishy bionic implant", kYellow },
    { "bumpy bionic implant", kBrightRed },
    { "glistening bionic implant", kYellow },
    { "spongy bionic implant", kBrightCyan },       // 10
    { "gelatinous bionic implant", kCyan },
    { "smelly bionic implant", kBrightBlue },
    { "slimy bionic implant", kWhite },
    { "bulbous bionic implant", kBrightMagenta },
    { "gossamer bionic implant", kGreen }           // 15
};

/* Floppy disks have their colors assigned by randomizeIlkNames ().
   Problem: When you write some program on a blank disk it changes color!
            Same if you reformat or degauss it. */
shObjectDescData FloppyData[30] =
{
    { "floppy disk labeled AUTOEXEC.BAT", kGray },
    { "floppy disk labeled AAIO", kGray },
    { "floppy disk labeled ABDA", kGray },
    { "floppy disk labeled BAYARD WENZEL", kGray },
    { "floppy disk labeled CONFIG.SYS", kGray },    //  5
    { "floppy disk labeled EIT ME", kGray },
    { "floppy disk labeled EUTOW", kGray },
    { "floppy disk labeled FORMAT C:", kGray },
    { "floppy disk labeled FWQWGADZ", kGray },
    { "floppy disk labeled GINOH DIVAD", kGray },   // 10
    { "floppy disk labeled IHTFP", kGray },
    { "floppy disk labeled JUSTIN BAILEY", kGray },
    { "floppy disk labeled LV FTS", kGray },
    { "floppy disk labeled PAPOU", kGray },
    { "floppy disk labeled PHAM NUWEN", kGray },    // 15
    { "floppy disk labeled RIBBET", kGray },
    { "floppy disk labeled SOO BAWLZ", kGray },
    { "floppy disk labeled THX 1138", kGray },
    { "floppy disk labeled XA 35", kGray },
    { "floppy disk labeled XERTH Q3", kGray },      // 20
    { "floppy disk labeled YERXA", kGray },
    { "floppy disk labeled ZARRO BOOGS", kGray },
#ifndef WIN32
    { "floppy disk labeled PRIME-"PRIME_VERSION".TAR.GZ", kGray }
#else
    { "floppy disk labeled PRIME-"PRIME_VERSION".ZIP", kGray }
#endif
};

shObjectDescData CanisterData[30] =
{
    { "aluminum canister", kCyan },
    { "brass canister", kBrown },
    { "copper canister", kRed },
    { "glass canister", kBrightCyan },
    { "grooved canister", kGreen },         //  5
    { "insulated canister", kBrightGreen },
    { "iron canister", kBrightBlue },
    { "magnesium canister", kWhite },
    { "plastic canister", kGray },
    { "pressurized canister", kBrightRed }, // 10
    { "reinforced canister", kYellow },
    { "square canister", kMagenta },
    { "steel canister", kCyan },
    { "tin canister", kBrown },
    { "titanium canister", kBrightCyan },   // 15
    { "vacuum sealed canister", kBrightBlue },
    { "squeezable canister", kMagenta },
    { "battered canister", kGray },
    { "chromed canister", kWhite },
    { "rusty canister", kRed },             // 20
    { "stained canister", kGreen },
    { "bulbous canister", kYellow }
};

shObjectDescData BeltData[5] =
{
    { "padded belt", kBrown },
    { "old belt", kBrightGreen },
    { "plated belt", kBrightBlue },
    { "heavy belt", kRed },
    { "rugged belt", kBrightRed }
};

shObjectDescData KeycardData[5] =
{
    { "red keycard", kRed },
    { "green keycard", kGreen },
    { "blue keycard", kBrightBlue },
    { "orange keycard", kBrightRed },
    { "purple keycard", kMagenta }
};

shObjectDescData GizmoData[5] =
{
    { "fiddly electronic gizmo", kBlue },
    { "gimmicky electronic gizmo", kBlue },
    { "fragile electronic gizmo", kBlue },
    { "complicated electronic gizmo", kBlue },
    { "weird electronic gizmo", kBlue }
};

/* Content generated by randomizeIlkNames. */
shObjectDescData GrenadeData[30];
shObjectDescData SGrenadeData[30];

void
randomizeIlkNames ()
{
    shuffle (CanisterData, 22, sizeof (shObjectDescData));
    shuffle (ImplantData, 15, sizeof (shObjectDescData));
    shuffle (FloppyData, 23, sizeof (shObjectDescData));
    shuffle (RayGunData, 12, sizeof (shObjectDescData));
    shuffle (JumpsuitData, 5, sizeof (shObjectDescData));
    shuffle (BeltData, 5, sizeof (shObjectDescData));
    shuffle (KeycardData, 5, sizeof (shObjectDescData));
    shuffle (GizmoData, 5, sizeof (shObjectDescData));

    /* Assign floppy colors to shuffled labels in a randomized
       repeating pattern. */
    static shColor floppyColors[] = {kRed, kGreen, kBrown, kMagenta,
        kCyan, kGray, kBrightRed, kBrightGreen, kYellow, kBrightBlue,
        kBrightMagenta, kBrightCyan, kWhite};
    const int numColors = sizeof (floppyColors) / sizeof (shColor);
    shuffle (floppyColors, numColors, sizeof (shColor));
    for (int i = 0; i < 23; ++i) {
        FloppyData[i].mColor = floppyColors[i % numColors];
    }

    /* Generate and assign grenade names. Prepare short names
       consisting only from third part. */

    /* Some data to cook names from.  Stylized after Xenocide. */
    static const char *adjs[] = {"heavy", "little", "shiny", "small",
        "strange", "tiny", "black", "brown", "gray", "red", "yellow", "white"};
    const int nadjs = sizeof (adjs) / sizeof (char *);
    static const char *mats[] = {"aluminum", "carbon-fiber", "electronic",
        "lead", "metal", "plastic", "plasteel", "sillicon", "steel",
        "titanium"};
    const int nmats = sizeof (mats) / sizeof (char *);
    static const char *shapes[] = {"ball", "box", "rounded box", "cube",
        "cuboid", "rounded cuboid", "cylinder", "device", "disc", "oval disc",
        "egg", "stick", "tetrahedron", "torus"};
    const int nshapes = sizeof (shapes) / sizeof (char *);
    const int tier2 = nadjs;
    const int tier3 = tier2 * nmats;

    int choices[30];
    for (int i = 0; i < 30; ++i) {
        bool ok;
        int n1, n2, n3;
        do { /* Make sure names do not repeat. */
            n1 = RNG (nadjs);
            n2 = RNG (nmats);
            n3 = RNG (nshapes);
            choices[i] = n1 + tier2 * n2 + tier3 * n3;
            ok = true;
            /* Check if new choice matches with any earlier. */
            for (int j = i-1; j >= 0; --j) {
                if (choices[j] == choices[i]) { /* It does! Retry. */
                    ok = false;
                    break;
                }
            }
        } while (!ok);
        snprintf (GrenadeData[i].mDesc, 40, "%s %s %s",
            adjs[n1], mats[n2], shapes[n3]);
        snprintf (SGrenadeData[i].mDesc, 40, "%s", shapes[n3]);
    }
}


shEnergyType
shObject::vulnerability ()
{
    if (isA (kEnergyCell) || (isA (kImplant) && mIlkId != kObjTorc)) {
        return kElectrical;
    } else if (isA (kCanister)) {
        if (isA (kObjLNO)) return kNoEnergy; /* Already frozen. */
        return kFreezing;
    } else if (isA (kObjGenericComputer) || isA (kFloppyDisk) ||
               isA (kObjGenericKeycard))
    {
        return kBugging;
    }
    switch (myIlk ()->mMaterial) {
    case kVegetable:
    case kFleshy:
    case kPaper:
    case kCloth:
    case kLeather:
    case kWood:
    case kPlastic:
    case kWax:
    case kBone:
        return kBurning;

    case kIron:
    case kBrass:
    case kTin:
    case kBronze:
    case kLead:
    case kSteel:
    case kAluminum:
    case kCopper:
    case kSilver:
    case kElectrum:
        return kCorrosive;

    case kSilicon:
    case kGlass:
    case kCrystal:
        return kElectrical;

    case kLiquid:
    case kTitanium:
    case kGold:
    case kPlatinum:
    case kMithril:
    case kGem:
    case kMineral:
    case kRock:
    case kCarbonFiber:
    case kPlasteel:
    case kAdamantium:
    case kEndurium:
    default:
        return kNoEnergy;
    }
}

void initializeObjCounts (); /* Defined later in this file. */

void
initializeObjects ()
{
    for (int i = 0; i < kMaxObjectType; ++i) {
        initGlyph (&ObjectGlyphs[i], ObjectSymbols[i], kBlack);
    }
    initializeMoney ();
    initializeTools ();
    initializeWeapons ();
    initializeRayGuns ();
    initializeArmor ();
    initializeCanisters ();
    initializeFloppyDisks ();
    initializeImplants ();
    initializeWreck ();
    initializeArtifacts ();
    initializeEnergy ();
    initializeLore ();
    initializeParser ();
    initializeObjCounts ();
    AllIlks[kObjNothing] = NULL;
}

int
shObjectIlk::isA (shObjId ilk)
{
    shObjId testilk = mId;
    while (kObjNothing != testilk) {
        if (testilk == ilk) {
            return 1;
        }
        testilk = AllIlks[testilk]->mParent;
    }
    return 0;
}

void
itemSpoilers ()
{
    FILE *spoilerfile = fopen ("items.txt", "w");

    for (int i = 0; i < 7; ++i) {
        int first, last;
        unsigned int total = 0;
        double prob;
        switch (i) {
            case 0:
                first = kObjFirstWeapon;
                last = kObjLastWeapon;
                break;
            case 1:
                first = kObjFirstArmor;
                last = kObjLastArmor;
                break;
            case 2:
                first = kObjFirstCanister;
                last = kObjLastCanister;
                break;
            case 3:
                first = kObjFirstFloppyDisk;
                last = kObjLastFloppyDisk;
                break;
            case 4:
                first = kObjFirstTool;
                last = kObjLastTool;
                break;
            case 5:
                first = kObjFirstRayGun;
                last = kObjLastRayGun;
                break;
            case 6:
                first = kObjFirstImplant;
                last = kObjLastImplant;
                break;
            default:
                first = kObjWreck;
                last = kObjWreck;
                break;
        }
        for (int k = first; k <= last; ++k) {
            total += AllIlks[k]->mProbability;
        }
        shObjectType type = AllIlks[first]->mType;
        fprintf (spoilerfile, "\n%s:\n", objectTypeHeader[type]);
        for (int k = first; k <= last; ++k) {
            shObjectIlk *ilk = AllIlks[k];
            prob = ilk->mProbability;
            if (prob > 0) prob = prob / total * 100;
            fprintf (spoilerfile, "%8.4f - %s\n", prob, ilk->mName);
        }
    }
    fclose (spoilerfile);
    I->p ("Tables saved in items.txt.");
}


void
shObject::name (const char *newname)
{
    if (!newname) {
        char prompt[80];
        char buf[80];

        snprintf (prompt, 80, "%same %s:", mUserName ? "Ren" : "N", these ());
        if (!I->getStr (buf, 40, prompt)) {
            I->p ("Never mind.");
            return;
        }
        newname = buf;
    }
    if (mUserName) {
      free ((void *) mUserName);
    }
    if (1 == strlen (newname) && ' ' == *newname) {
        mUserName = NULL;
    } else {
        mUserName = strdup (newname);
    }
}


void
shObject::nameIlk ()
{
    char prompt[80];
    char buf[80];
    const char *tmp;
    const char *desc;

    tmp = myIlk ()->mUserName;
    myIlk ()->mUserName = NULL;
    desc = getDescription ();
    myIlk ()->mUserName = tmp;

    snprintf (prompt, 80, "Call %s:", desc);
    if (!I->getStr (buf, 40, prompt)) {
        I->p ("Never mind.");
        return;
    }
    if (myIlk ()->mUserName) {
      free ((void *) myIlk ()->mUserName);
      myIlk ()->mUserName = NULL;
    }
    if (1 == strlen (buf) && ' ' == *buf) {
        return;
    } else {
        myIlk ()->mUserName = strdup (buf);
    }
}


const char *
shObject::getVagueDescription ()
{
    return myIlk ()->mVagueName;
}


const char *
shObject::getShortDescription ()
{
    int l = 0;
    char *buf = GetBuf ();
    const char *name;
    name = !isAppearanceKnown () ? myIlk ()->mVagueName :
            isIlkKnown () ? myIlk ()->mName : myIlk ()->mAppearance;

    if (myIlk ()->mUserName) {
        l += snprintf (buf + l, SHBUFLEN - l, "%s", name);
        if (mCount > 1) {
            makePlural (buf, SHBUFLEN);
            ++l;
        }
        l += snprintf (buf + l, SHBUFLEN - l, " called %s", myIlk ()->mUserName);
    } else {
        l += snprintf (buf + l, SHBUFLEN - l, "%s", name);
        if (mCount > 1) {
            makePlural (buf, SHBUFLEN);
            ++l;
        }
    }
    return buf;
}


const char *
shObject::getDescription ()
{
    int l = 0;
    const char *fooproof = NULL;
    const char *dmg;
    char *buf = GetBuf ();

    if (isBugginessKnown () && !isBugProof ()) {
        if (mIlkId != kObjTorc) {
            l += snprintf (buf + l, SHBUFLEN - l, "%s",
                           isBuggy () ? "buggy " :
                           isOptimized () ? "optimized " : "debugged ");
        } else {
            l += snprintf (buf + l, SHBUFLEN - l, "%s",
                           isBuggy () ? "gray " :
                           isOptimized () ? "golden " : "silver ");
        }
    }

    switch (vulnerability ()) {
    case kCorrosive: fooproof = "rustproof"; dmg = "corroded"; break;
    case kBurning: fooproof = "fireproof"; dmg = "burnt"; break;
    case kBugging: fooproof = "protected"; dmg = "infected"; break;
    default: dmg = "damaged";
    }

    /* Computers and floppy disks hide infectedness by default. */
    if (isDamageKnown ()) {
        switch (mDamage) {
        case 1:
            l += snprintf (buf + l, SHBUFLEN - l, "%s ", dmg); break;
        case 2:
            l += snprintf (buf + l, SHBUFLEN - l, "very %s ", dmg); break;
        case 3:
            l += snprintf (buf + l, SHBUFLEN - l, "extremely %s ", dmg); break;
        default: break;
        }
    }

    if (isFooproof () && isFooproofKnown ()) {
        l += snprintf (buf + l, SHBUFLEN - l, "%s ", fooproof);
    }

    if (isCracked () && isCrackedKnown ()) {
        l += snprintf (buf + l, SHBUFLEN - l, "cracked ");
    }

    if (isEnhanceable () && isEnhancementKnown ()) {
        l += snprintf (buf + l, SHBUFLEN - l, "%s%d ",
                       mEnhancement < 0 ? "" : "+", mEnhancement);
    }

    if (kObjWreck == mIlkId) {
        l += snprintf (buf + l, SHBUFLEN - l, "disabled %s", MonIlks[mWreckIlk]->mName);
        if (mCount > 1) {
            makePlural (buf, SHBUFLEN);
            ++l;
        }
    } else {
        if (!isAppearanceKnown ()) {
            l += snprintf (buf + l, SHBUFLEN - l, "%s", myIlk ()->mVagueName);
            if (mCount > 1) {
                makePlural (buf, SHBUFLEN);
                ++l;
            }
        } else {
            l += snprintf (buf + l, SHBUFLEN - l, "%s",
                           isIlkKnown () ? myIlk ()->mName : myIlk ()->mAppearance);
            if (mCount > 1) {
                makePlural (buf, SHBUFLEN);
                ++l;
            }
        }
        if (myIlk ()->mUserName) {
            l += snprintf (buf + l, SHBUFLEN - l, " called %s", myIlk ()->mUserName);
        }
    }
    if (mUserName) {
        l += snprintf (buf + l, SHBUFLEN - l, " named %s", mUserName);
    }

    if (isChargeable () && isChargeKnown () && !isA (kObjEmptyRayGun)) {
        l += snprintf (buf + l, SHBUFLEN - 1, " (%d charg%s)", mCharges,
                       mCharges == 1 ? "e" : "es");
    }
    buf[SHBUFLEN-1] = 0;
    return buf;
}


const char *
shObject::inv ()
{
    char *buf = GetBuf ();
    int l;

    if (mCount > 1) {
        l = snprintf (buf, SHBUFLEN, "%d %s", mCount, getDescription ());
    } else {
        const char *tmp = getDescription ();
        l = snprintf (buf, SHBUFLEN, "%s %s",
                      isUniqueKnown () ? "the" :
                      isvowel (tmp[0]) ? "an" : "a", tmp);
    }
    if (isWorn ()) {
        if (isA (kImplant)) {
            l += snprintf (buf + l, SHBUFLEN - l, " (implanted in %s)",
                           describeImplantSite (mImplantSite));
        } else {
            l += snprintf (buf + l, SHBUFLEN - l, " (worn");
            if (isA (kObjBioMask) && isIlkKnown ()) {
                l += snprintf (buf + l, SHBUFLEN - l, ", %s",
                               isToggled () ? "EM field mode" : "thermal mode");
            }
            l += snprintf (buf + l, SHBUFLEN - l, ")");
        }
    } else if (isWielded ()) {
        l += snprintf (buf + l, SHBUFLEN - l, " (wielded");
        if (isSelectiveFireWeapon () && isIlkKnown ()) {
            l += snprintf (buf + l, SHBUFLEN - l, ", %s",
                           isToggled () ? "burst mode" : "single fire");
        }
        l += snprintf (buf + l, SHBUFLEN - l, ")");
    }
    if (isDesignated ()) {
        l += snprintf (buf + l, SHBUFLEN - l, " (designated)");
    }
    if (isActive ()) {
        l += snprintf (buf + l, SHBUFLEN - 1, " (activated)");
        if (isA (kObjGeigerCounter)) {
            if (Hero.getStoryFlag ("geiger counter message")) {
                l += snprintf (buf + l, SHBUFLEN - 1, " (clicking)");
            }
        }
    }
    if (isUnpaid ()) {
        l += snprintf (buf + l, SHBUFLEN - 1, " (unpaid, $%d)",
                       mCount * myIlk ()->mCost);
    } else if (Hero.isInShop () && Level->getShopKeeper (Hero.mX, Hero.mY)) {
        l += snprintf (buf + l, SHBUFLEN - 1, " (worth $%d)",
                       mCount * myIlk ()->mCost);
    }
/*  if (GodMode) {
        l += snprintf (buf + l, SHBUFLEN - l, " (%d g)", getMass ());
    }*/
    buf[SHBUFLEN-1] = 0;
    return buf;
}


const char *
shObject::her (shCreature *owner)
{
    char *buf = GetBuf ();
    const char *pronoun = "its";
    if (kFemale == owner->mGender)
        pronoun = "her";
    else if (kMale == owner->mGender)
        pronoun = "his";
    if (mCount > 1) {
        snprintf (buf, SHBUFLEN, "%s %d %s",
                      pronoun, mCount, getDescription ());
    } else {
        snprintf (buf, SHBUFLEN, "%s %s", pronoun, getDescription ());
    }
    return buf;
}


int
shObject::isA (shObjId type)
{
    return myIlk ()->isA (type);
}


int
shObject::canMerge (shObject *obj)
{
    if (!obj) {
        obj = this;
    }
    if (obj->myIlk () != myIlk () || AllIlks[kObjWreck] == myIlk ()) {
        return 0;
    }
    if ((myIlk ()->mFlags & kMergeable) &&
        (obj->mFlags == mFlags) &&
        (obj->mBugginess == mBugginess) &&
        (obj->mDamage == mDamage) &&
        (!obj->mUserName && !mUserName))
    {
        return 1;
    }
    return 0;
}


static shObjectVector DeletedObjects;

void
shObject::merge (shObject *obj)
{
    if (obj == this) return;
    mCount += obj->mCount;
    /* HACK: don't delete the object right away; we might need to use
       it to print a message or something.  (For example, the case of
       a shopkeeper giving a quote about an object picked up in a
       shop.) */
    DeletedObjects.add (obj);
    //delete obj;
}


shObject *
shObject::split (int count)
{
    shObject *result;

    assert (count < mCount);

    mCount -= count;
    result = new shObject ();
    result->mIlkId = mIlkId;
    result->mCount = count;
    result->mHP = mHP;
    result->mBugginess = mBugginess;
    result->mFlags = mFlags;
    result->mWreckIlk = mWreckIlk;
    return result;
}


/* returns number of objects destroyed from stack */
int
shObject::sufferDamage (shAttack *attack, int x, int y,
                        int multiplier, int specialonly)
{
    if (isIndestructible ()) return 0;
    int damage;
    int totaldamage = 0;
    int annihilated = 0;
    int cansee = Hero.canSee (x, y);
    int numdestroyed = RNG (1, 2);
    int nearby = distance (&Hero, x, y) <= 30;
    const char *theobj;

    if (Hero.owns (this)) {
        cansee = 1;
        theobj = your ();
    } else if (mOwner) {
        theobj = her (mOwner);
    } else {
        theobj = theQuick ();
    }

    if (numdestroyed > mCount) {
        numdestroyed = mCount;
    }

    for (int j = 0; j < 3; ++j) {
        shEnergyType energy = (shEnergyType) attack->mDamage[j].mEnergy;

        if (kNoEnergy == energy) {
            break;
        } else if (kCorrosive == energy) {
            if (!Hero.owns (this))
                cansee = 0; /* reduce messages */
            if (isA (kEnergyCell)) {
                if (cansee) {
                    if (1 == mCount) {
                        I->p ("%s dissolves!", theobj);
                    } else {
                        I->p ("%s dissolve!", theobj);
                    }
                }
                return mCount;
            } else if (kCorrosive == vulnerability ()) {
                if (isFooproof ()) {
                    if (cansee) {
                        setFooproofKnown ();
                        I->p ("%s is not affected.", theobj);
                    }
                    continue;
                }
                if (isOptimized () && RNG (4)) {
                    if (cansee) {
                        setBugginessKnown ();
                        I->p ("Somehow, %s is not affected.", theobj);
                    }
                    continue;
                }
                if (mDamage < 3)
                    ++mDamage;
                debug.log ("corroded %s %d", theobj, mDamage);
                if (cansee) {
                    switch (mDamage) {
                    case 1:
                        I->p ("%s corrodes!", theobj); break;
                    case 2:
                        I->p ("%s corrodes some more!", theobj); break;
                    default:
                        I->p ("%s is thoroughly corroded!", theobj); break;
#if 0               /* this makes it too easy to dispose of a welded weapon */
                    default:
                        I->p ("%s dissolves completely!", theobj);
                        return 1;
#endif
                    }
                }
                setFooproofKnown (); /* Obviously it got damaged. */
            } else if (cansee) {
                I->p ("%s is not affected.", theobj);
            }
        } else if (kElectrical == energy) {
            if (kElectrical == vulnerability ()) {
                damage = multiplier *
                    RNG (attack->mDamage[j].mLow, attack->mDamage[j].mHigh);
                totaldamage += damage;
                totaldamage -= myIlk ()->mHardness;
                if (totaldamage > mHP) {
                    if (cansee) {
                        if (Hero.owns (this) && isWorn ()) {
                            I->p ("%s is ejected from your %s in a shower of "
                                  "sparks!",
                                  theobj, describeImplantSite (mImplantSite));
                            Hero.doff (this);
                            return 0;
                        }
                    }
                    return 1;
                } else {
                    /* either it explodes or not, no damage */
                    totaldamage = 0;
                }
            }
        } else if (kBurning == energy) {
            if (isA (kMoney)) {
                if (1 == mCount) {
                    I->p ("%s melts!", theobj);
                } else {
                    I->p ("%s melt!", theobj);
                }
                return mCount;
            } else if (isA (kFloppyDisk)) {
                if (cansee) {
                    if (1 == mCount) {
                        I->p ("%s melts!", theobj);
                    } else if (1 == numdestroyed) {
                        I->p ("One of %s melts!", theobj);
                    } else if (numdestroyed == mCount) {
                        I->p ("%s melt!", theobj);
                    } else {
                        I->p ("Some of %s melt!", theobj);
                    }
                } else if (nearby) { /* FIXME: no msg needed when underwater*/
                    I->p ("You smell burning plastic.");
                }
                return numdestroyed;
            } else if (kBurning == vulnerability ()) {
                if (!Hero.owns (this))
                    cansee = 0; /* reduce messages */
                if (isFooproof ()) {
                    if (cansee) {
                        setFooproofKnown ();
                        I->p ("%s is not affected.", theobj);
                    }
                    continue;
                }
                if (isOptimized () && RNG (4)) {
                    if (cansee) {
                        setBugginessKnown ();
                        I->p ("Somehow, %s is not affected.", theobj);
                    }
                    continue;
                }
                if (mDamage < 4)
                    ++mDamage;
                debug.log ("burned %s %d", theobj, mDamage);
                if (cansee) {
                    switch (mDamage) {
                    case 1:
                        I->p ("%s burns!", theobj); break;
                    case 2:
                        I->p ("%s burns some more!", theobj); break;
                    case 3:
                        I->p ("%s is thoroughly burned!", theobj); break;
                    default:
                        I->p ("%s burns completely!", theobj);
                    }
                }
                setFooproofKnown ();
                if (mDamage == 4) return 1;
            }
        } else if (kFreezing == energy) {
            if (isA (kCanister) && kFreezing == vulnerability ()) {
                if (cansee) {
                    if (1 == mCount ) {
                        I->p ("%s freezes and shatters!", theobj);
                    } else if (1 == numdestroyed) {
                        I->p ("One of %s freezes and shatters!", theobj);
                    } else if (numdestroyed == mCount) {
                        I->p ("%s freeze and shatter!", theobj);
                    } else {
                        I->p ("Some of %s freeze and shatter!", theobj);
                    }
                } else if (nearby) { /* FIXME: no msg when in vacuum */
                    I->p ("You hear something shatter.");
                }
                return numdestroyed;
            }
        } else if (kMagnetic == energy) {
            damage = RNG (attack->mDamage[j].mLow, attack->mDamage[j].mHigh);
            if ((isA (kObjM56Smartgun) || isA (kObjM57Smartgun)) &&
                (25 <= damage))
            {   /* Erase auto-aim system. */
                mIlkId = kObjPulseRifle;
                if (mOwner == &Hero) Hero.computeSkills ();
            }
            /* FIXME: clerkbot should get mad if this is hero's fault */
            if (isComputer ()) { /*  */
                if (RNG (50) < damage) {
                    mEnhancement = -5; /* lost OS */
                }
            } else {
                /* Introduce some variety in damage. It would be boring if
                   a hit either damaged no equipment or all for -1 enhancement.
                   Similarly neutralizing bugginess for every item or none is
                   undesirable. */
                int fracloss = mini (maxi (5 + RNG (3) - damage / 3, 0), 5);
                mEnhancement = mEnhancement * fracloss / 5;
            }
            if (RNG (50) < damage) {/* Neutralizing bugginess is tough. */
                mBugginess = 0;
            }
            if (RNG (12) < damage && isA (kFloppyDisk) &&
                !(isA (kObjBlankDisk)))
            {   /* Floppy disks easily lose information. */
                mIlkId = kObjBlankDisk;
                resetFooproof ();
                setFooproofKnown ();
                resetInfected ();
                setInfectedKnown ();
            }
        } else if (kBugging == energy) {
            if (isOptimized ()) {
                setDebugged ();
            } else {
                setBuggy ();
            }
            if (isA (kObjGenericComputer)) {
                if (!isFooproof () && !RNG (3)) {
                    mDamage = 1;    /* infect! */
                }
            }
        } else if (kPoisonous == energy) {
            if (isA (kObjBioArmor) && !isA (kObjMBioArmor0)) {
                if (!Hero.owns (this) &&
                    (Hero.isBlind () || x != Hero.mX || y != Hero.mY))
                {
                    cansee = 0;
                }
                if (isOptimized () && RNG (4)) {
                    if (cansee) {
                        I->p ("Somehow, %s is not affected.", theobj);
                    }
                    continue;
                }
                --mEnhancement;
                debug.log ("poisoned %s %d", theobj, mEnhancement);
                if (cansee) {
                    if (mEnhancement == -4) {
                        I->p ("%s quivers and dies!", theobj);
                    } else {
                        I->p ("%s weakens!", theobj);
                    }
                }
                if (mEnhancement == -4) return 1;
            } else if (isA (kObjGreenCPA)) {
                if (!Hero.owns (this)) {
                    cansee = 0;
                }
                int worked = 0;
                if (mEnhancement < 3) {
                    ++mEnhancement;
                    worked = 1;
                }
                debug.log ("poisoned %s %d", theobj, mEnhancement);
                if (cansee && worked) {
                    if (!Hero.isBlind ()) {
                        I->p ("%s glistens.", theobj);
                    } else {
                        I->p ("%s throbs.", theobj);
                    }
                }
            }
        } else if (kPlaguing == energy) {
            if (isA (kObjBioArmor) && !isA (kObjMBioArmor1)) {
                int domsg = 1;
                damage = 3;
                if (!Hero.owns (this)) {
                    cansee = 0;
                }
                if (isOptimized () && RNG (4)) {
                    if (cansee) {
                        if (mEnhancement > -3) {
                            I->p ("%s manages to resist somewhat.", theobj);
                        } else {
                            I->p ("%s tries to resist but fails.", theobj);
                        }
                    }
                    damage = 1;
                    domsg = 0;
                }
                mEnhancement -= damage;
                debug.log ("plagued %s %d", theobj, mEnhancement);
                if (domsg) {
                    if (mEnhancement <= -4) {
                        I->p ("%s gets horribly sick and dies!", theobj);
                    } else {
                        I->p ("%s gets much weaker!", theobj);
                    }
                }
                if (mEnhancement <= -4) return 1;
            }
        } else if (kHealing == energy) {
            if (isA (kObjBioArmor)) {
                if (!Hero.owns (this) &&
                    (Hero.isBlind () || x != Hero.mX || y != Hero.mY))
                {
                    cansee = 0;
                }
                int worked = 0;
                if (mEnhancement < 2) {
                    ++mEnhancement;
                    worked = 1;
                }
                debug.log ("healed %s %d", theobj, mEnhancement);
                if (cansee) {
                    if (worked) {
                        I->p ("%s feels tougher!", theobj);
                    } else {
                        I->p ("%s seems to be enjoying something.", theobj);
                    }
                }
            }
        } else if (kRestoring == energy) {
            if (isA (kObjBioArmor)) {
                if (!Hero.owns (this) &&
                    (Hero.isBlind () || x != Hero.mX || y != Hero.mY))
                {
                    cansee = 0;
                }
                int worked = 0;
                if (isBuggy ()) {
                    setDebugged ();
                    worked += 1;
                }
                if (mEnhancement < 0) {
                    mEnhancement = 0;
                    worked += 1;
                }
                if (mDamage) {
                    mDamage = 0;
                    worked += 1;
                }
                debug.log ("restored %s %d", theobj, mEnhancement);
                if (cansee) {
                    switch (worked) {
                    case 0: I->p ("%s seems to be enjoying something.", theobj);
                        break;
                    case 1: I->p ("%s feels better!", theobj); break;
                    case 2: I->p ("%s feels much better!", theobj); break;
                    case 3: I->p ("%s feels great!", theobj); break;
                    }
                }
            }
        } else if (kDisintegrating == energy) {
            /* Rays travel above objects. */
            if (attack->mType != shAttack::kDisintegrationRay) {
                totaldamage += 500; /* Otherwise annihilation. */
                annihilated = 1;
            }
        } else if (specialonly) {
            continue;
        } else if (kConcussive == energy && kLeather == myIlk ()->mMaterial) {
            continue;
        } else {
            damage = multiplier *
                RNG (attack->mDamage[j].mLow, attack->mDamage[j].mHigh);
            totaldamage += damage;
        }
        multiplier = 1;
    }

    debug.log ("dealt %d damage to object %p", totaldamage, this);

    totaldamage -= myIlk ()->mHardness;
    if (totaldamage <= 0) {
        return 0;
    }
    mHP -= totaldamage;
    if (mHP <= 0) {
        mHP = 0;
        if (cansee) {
            if (annihilated) {
                I->p ("%s is annihilated!", theobj);
            } else if (myIlk ()->mMaterial == kGlass) {
                I->p ("%s shatters!", theobj);
            } else {
                I->p ("%s is destroyed!", theobj);
            }
        } else if (nearby && myIlk ()->mMaterial == kGlass) {
            I->p ("You hear something shatter.");
        }
        return 1;
    }
    if (isA (kObjGenericComputer)) {
        /* Computers generally do not like to be kicked around... */
        if (!isUpdateTime (mEnhancement)) {
            I->p ("Sparks fly from %s!", theobj);
            setBuggy ();
        /* ... but during update process this is acceptable. */
        } else {
            resetUpdatePeriod ();
        }
    }
    return 0;
}


int
shObject::getPsiModifier ()
{
    if (isA (kObjHarmonyDome) && !isBuggy ()) {
        /* Negate all psi aura disruption caused by cranial implants. */
        int psimod = 0;
        for (int i = 0; i <= shObjectIlk::kCerebellum; ++i) {
            if (mOwner->mImplants[i] &&
                mOwner->mImplants[i]->myIlk ()->mPsiModifier < 0)
            {
                psimod -= mOwner->mImplants[i]->myIlk ()->mPsiModifier;
            }
        }
        return psimod;
    } else if (isA (kObjPsiAmp) || isA (kObjSBioArmor4)) {
        return mEnhancement + myIlk ()->mPsiModifier;
    } else {
        return myIlk ()->mPsiModifier;
    }
}


int
shObject::getArmorBonus ()
{
    if (kArmor != myIlk ()->mType) {
        return 0;
    } else {
        return maxi (0, myIlk ()->mArmorBonus - mDamage + mEnhancement);
    }
}


int
shObject::getThreatRange (shCreature *target)
{
    if (isA (kWeapon)) {
        return 20;
    } else { /* impossible to critical hit with improvised weapon */
        return 999;
    }
}

int
shObject::getCriticalMultiplier ()
{
    if (isA (kWeapon)) {
        return 2;
    } else { /* impossible to critical hit with improvised weapon */
        return 1;
    }
}


static struct shProbTotals
{
    int mArmSum;
    int mWpnSum;
    int mRaySum;
    int mCanSum;
    int mImpSum;
    int mDiskSum;
    int mToolSum;
} totals;


static shObject *
pickFromRange (shObjId from, shObjId to, int total)
{
    if (from > to) {
        return NULL;
    }

    int r = RNG (total);
    int n = 0;
    for (int i = from; i <= to; ++i) {
        int p = AllIlks[i]->mProbability;
        if (p == ABSTRACT) continue;
        n += p;
        if (n >= r) {
            return new shObject ((shObjId) i);
        }
    }
    debug.log ("Failed pick an ilk! [%d,%d] (%d)", from, to, total);
    return NULL;
}


shObject *
createWeapon ()
{
    return pickFromRange (kObjFirstWeapon, kObjLastWeapon, totals.mWpnSum);
}

shObject *
createArmor ()
{
    return pickFromRange (kObjFirstArmor, kObjLastArmor, totals.mArmSum);
}

shObject *
createTool ()
{
    return pickFromRange (kObjFirstTool, kObjLastTool, totals.mToolSum);
}

shObject *
createCanister ()
{
    return pickFromRange (kObjFirstCanister, kObjLastCanister, totals.mCanSum);
}

shObject *
createFloppyDisk ()
{
    return pickFromRange (kObjFirstFloppyDisk, kObjLastFloppyDisk, totals.mDiskSum);
}

shObject *
createImplant ()
{
    return pickFromRange (kObjFirstImplant, kObjLastImplant, totals.mImpSum);
}

shObject *
createRayGun ()
{
    return pickFromRange (kObjFirstRayGun, kObjLastRayGun, totals.mRaySum);
}

/* generate random object according to level specification:
   -1 level indicates shop
*/
shObject *
generateObject (int level)
{   /* first what kind of object? */
    while (1) {
        int type = RNG (100);
        if (type < 20) {
            if (-1 != level) {
                /* 10d50 is largest amount possible. */
                shObject *cash = new shObject (kObjMoney);
                cash->mCount =
                    (NDX (mini (level, 10), 20 + 3 * mini (level, 10)));
                return cash;
            }
        } else if (type < 35) {
            return createWeapon ();
        } else if (type < 45) {
            return createArmor ();
        } else if (type < 48) {
            return new shObject (kObjEnergyCell);
        } else if (type < 60) {
            return createTool ();
        } else if (type < 75) {
            return createCanister ();
        } else if (type < 93) {
            return createFloppyDisk ();
        } else if (type < 98) {
            return createImplant ();
        } else if (type < 100) {
            return createRayGun ();
        } else if (-1 == level) {
            continue;
        } else {
            return new shObject (kObjMoney);
        }
    }
    return NULL; /* Not reached. */
}


static int
sumProb (shObjId from, shObjId to)
{
    int sum = 0;
    for (int i = from; i <= to; ++i)
        if (AllIlks[i]->mProbability != ABSTRACT)
            sum += AllIlks[i]->mProbability;
    return sum;
}


void
initializeObjCounts ()
{
    totals.mArmSum = sumProb (kObjFirstArmor, kObjLastArmor);
    totals.mWpnSum = sumProb (kObjFirstWeapon, kObjLastWeapon);
    totals.mRaySum = sumProb (kObjFirstRayGun, kObjLastRayGun);
    totals.mCanSum = sumProb (kObjFirstCanister, kObjLastCanister);
    totals.mImpSum = sumProb (kObjFirstImplant, kObjLastImplant);
    totals.mDiskSum = sumProb (kObjFirstFloppyDisk, kObjLastFloppyDisk);
    totals.mToolSum = sumProb (kObjFirstTool, kObjLastTool);
}


static int InventorySortOrder [kMaxObjectType] =
{
    11, 0, 5, 8, 6, 11, 3, 1, 2, 10, 7, -1
};

/* used for sorting
   returns: -1, 0 or 1 depending on whether obj1 belongs before, with, or
            after obj2 in an inventory list.  sorts on type
*/
int
compareObjects (shObject **obj1, shObject **obj2)
{
    int t1 = InventorySortOrder[(*obj1)->myIlk ()->mType];
    int t2 = InventorySortOrder[(*obj2)->myIlk ()->mType];
    if (t1 < t2) return -1;
    if (t1 > t2) return 1;
    return 0;
}


void
purgeDeletedObjects ()
{
    while (DeletedObjects.count ()) {
        delete DeletedObjects.removeByIndex(0);
    }
}

/* Returns object that should be shown at the top of the stack. */
shObject *   /* This means: find most interesting thing to player. */
findBestObject (shObjectVector *objs)
{   /* Finds object with the lowest objecttype. */
    shObject *bestobj = NULL;
    int besttype = kMaxObjectType;

    for (int i = 0; i < objs->count (); ++i) {
        if (objs->get (i)->myIlk ()->mType < besttype) {
            besttype = objs->get (i)->myIlk ()->mType;
            bestobj = objs->get (i);
        }
    }
    return bestobj;
}


shObject::shObject (shObjId id)
{
    mIlkId = id;
    mFlags = 0;
    mLetter = 0;
    mUserName = NULL;
    mOwner = NULL;
    mLastEnergyBill = MAXTIME;

    /* Set count. */
    if (myIlk ()->mFlags & kMergeable) {
        /* Ammunition. Bullets, shells, slugs, fuel drums and like. */
        if (isA (kProjectile)) {
            int range = (int) (18.0 / sqrt ((double) myIlk ()->mWeight));
            mCount = RNG (range, range * 10);
        } /* Shurikens, daggers and grenades. */
        else if (myIlk ()->mFlags & kMissile) mCount = RNG (1, 4);
        else if (id == kObjEnergyCell) mCount = RNG (8, 80);
        /* Canisters can merge but are usually generated single. */
        else mCount = 1;
    } else { /* Guns, armor pieces, tools etc. */
        mCount = 1;
    }

    /* Set bugginess. */
    if (kBugProof & myIlk ()->mFlags) {
        mBugginess = 0;
        setBugginessKnown ();
    } else if (kUsuallyBuggy & myIlk ()->mFlags && RNG (9)) {
        mBugginess = -1;
    } else {
        int tmp;
        if (isA (kImplant) || isA (kFloppyDisk)) tmp = RNG (6);
        else tmp = RNG (8);
        mBugginess = (1 == tmp) ? 1 : (0 == tmp) ? -1 : 0;
    } /* Special cases for some items. */
    if (kObjSpamDisk == mIlkId && RNG (4)) {
        mBugginess = -1;
    } else if (kObjBugDetectionDisk == mIlkId && RNG (2)) {
        mBugginess = 1;
    }

    /* Set fooproofnes or crackedness. */
    if (isA (kImplant)) {
        setFooproofKnown (); /* No insulated fooproofness for implants yet. */
    } else if (isA (kCanister)) {
        setFooproofKnown ();
    } else if (isA (kFloppyDisk)) {
        if (!RNG (100) &&
            mIlkId != kObjHackingDisk && mIlkId != kObjMatterCompiler)
        {
            setCracked ();
        }
    } else if (isA (kObjGenericKeycard)) {
        if (!RNG (100) && !isA (kObjMasterKeycard)) setCracked ();
    } else if (isA (kObjGenericComputer)) {
    } else if (!isA (kWeapon) && !isA (kArmor)) {
        setFooproofKnown ();
    } else if (kNoEnergy != vulnerability ()) {
    } else { /* Cannot have fooproofness anyway. */
        setFooproofKnown ();
    }

    /* Set enhancement. Requires bugginess determined beforehand. */
    if (!isEnhanceable ()) {
        mEnhancement = 0;
    } else if (isA (kArmor)) {
        int max = myIlk ()->mMaxEnhancement;
        if (isOptimized ()) {
            mEnhancement = (RNG (max) * RNG (max) + max - 1) / max;
        } else if (isBuggy ()) {
            mEnhancement = RNG (max) * RNG (max) / -max;
        } else {
            mEnhancement = RNG (6) ? RNG (max) * RNG (max) / max
                                   : RNG (max) * RNG (max) / -max;
        }
    } else if (isA (kWeapon)) {
        int tmp = RNG (20); /* -1 or +1 weapons are rare. */
        mEnhancement = (1 == tmp) ? 1 : (0 == tmp) ? -1 : 0;
    } else if (isA (kImplant)) {
        if (isOptimized ()) {
            mEnhancement = RNG (1, 5);
        } else if (isBuggy ()) {
            mEnhancement = -RNG (1, 5);
        } else {
            mEnhancement = RNG (6) ? RNG (0, 2) + RNG (0, 3) : -RNG (0, 5);
        }
    } else if (isA (kObjGenericComputer)) {
        if (!RNG (10)) {
            mEnhancement = -5; /* No OS! */
        } else if (!RNG (8)) {
            mEnhancement = RNG(-4, +4); /* Hardware independent OS. */
        } else switch (mBugginess) { /* Some OS. */
            case -1: mEnhancement = RNG(-4, -1); break;
            case  0: mEnhancement = RNG(-1, +1); break;
            case +1: mEnhancement = RNG(+1, +4); break;
        }
    } else { /* Other enhanceable items start out without it. */
        mEnhancement = 0;
    }

    /* Set charges. */
    if (mIlkId == kObjEmptyRayGun) {
        mCharges = 0;
    } else if (isA (kRayGun)) {
        mCharges = NDX (2, 6);
    } else if (isA (kObjGenericPlant)) {
        mCharges = RNG (1, myIlk ()->mMaxCharges);
        setChargeKnown ();
    } else if (isA (kObjGenericEnergyTank)) {
        mCharges = RNG (0, myIlk ()->mMaxCharges / 5);
        setChargeKnown ();
    } else if (mIlkId == kObjKhaydarin) {
        mCharges = NDX (2, 2);
    } else if (mIlkId == kObjPlasmaCaster) {
        mCharges = !RNG (10) ? 1 : 0;
    } else {
        mCharges = RNG (0, myIlk ()->mMaxCharges);
    }

    /* Set damage or infectedness. */
    mDamage = 0; /* Currently no items start damaged. */
    if (!isA (kFloppyDisk) && !isComputer ()) {
        /* Floppies and computers may be infected (damaged) */
        setDamageKnown ();
    } else if (isComputer ()) {
    } else if (mIlkId == kObjBlankDisk) {
        setInfectedKnown ();
        setCrackedKnown ();
    } else if (mIlkId == kObjComputerVirus ||
               (mIlkId != kObjAntivirus && !RNG (20)))
    {
        setInfected ();
    }

    /* Set toggledness. */
    if (isSelectiveFireWeapon ()) {
        setToggled (); /* Guns start out in burst mode. */
    } else if (isA (kObjBioMask)) {
        if (RNG (2)) setToggled (); /* Random mode. */
    }

    /* Remaining properties. */
    mHP = myIlk ()->mHP;
    if (!isEnhanceable ())
        setEnhancementKnown ();
    if (!isChargeable ())
        setChargeKnown ();
    if (!isA (kFloppyDisk) && !isA (kObjGenericKeycard))
        setCrackedKnown ();

    if (!strcmp (myIlk ()->mVagueName, myIlk ()->mAppearance) &&
        !isA (kObjTorc))
    {
        setAppearanceKnown ();
    }
}
